@@ -14,16 +14,15 @@
 
 """Core brax structs and some conversion and slicing functions."""
 
-import os
+import copy
 from typing import Optional, Sequence, Tuple, Union
 import warnings
 
 from brax import jumpy as jp
 from brax import math
-from brax.io import file
+from brax.io import mesh
 from brax.physics import config_pb2
 from flax import struct
-from trimesh.exchange.load import load_mesh
 
 
 @struct.dataclass
@@ -142,16 +141,24 @@ class Info(object):
     contact: External contact forces applied at a step
     joint: Joint constraint forces applied at a step
     actuator: Actuator forces applied at a step
+    contact_pos: Contact point position in world coordinates.
+    contact_normal: Contact normal.
+    contact_penetration: Contact penetration.
   """
   contact: P
   joint: Union[P, Q]
   actuator: P
+  contact_pos: jp.ndarray
+  contact_normal: jp.ndarray
+  contact_penetration: jp.ndarray
 
 
 def validate_config(
     config: config_pb2.Config,
     resource_paths: Optional[Sequence[str]] = None) -> config_pb2.Config:
   """Validate and normalize config settings for use in systems."""
+  config = copy.deepcopy(config)
+
   if config.dt <= 0:
     raise ValueError('config.dt must be positive')
 
@@ -171,48 +178,35 @@ def validate_config(
   find_dupes(config.mesh_geometries)
 
   if config.dynamics_mode == 'legacy_spring':
-    for j in config.joints:
-      if j.stiffness == 0:
-        raise ValueError(
-            f'Joint {j.name} has 0 stiffness, but you have specified the legacy dynamics mode.  This may cause unexpected behavior.'
-        )
-  else:
-    if config.dynamics_mode != 'pbd':
-      warnings.warn(
-          """Dynamics mode either not specified or not recognized, defaulting to position based.  If you wish to preserve legacy behavior used in previous versions of Brax, simply append `dynamics_mode: "legacy_spring"` to your config (without '`'s)."""
-      )
-    if config.baumgarte_erp:
+    if any(j.stiffness == 0 for j in config.joints):
       raise ValueError(
-          'Baumgarte_erp specified by config, but is not compatible with the pbd dynamics mode.'
-      )
+          'joint.stiffness must be >0 when dynamics_mode == legacy_spring')
+  elif config.dynamics_mode == 'pbd':
+    if any(j.stiffness != 0 for j in config.joints):
+      raise ValueError('joint.stiffness is invalid when dynamics_mode == pbd')
+    if config.baumgarte_erp:
+      raise ValueError('baumgarte_erp is invalid when dynamics_mode == pbd')
+  elif any(j.stiffness != 0 for j in config.joints):
+    config.dynamics_mode = 'legacy_spring'
+    warnings.warn('dynamics_mode not specified, but joint.stiffness >0. '
+                  'Setting dynamics_mode="legacy_spring".')
+  else:
     config.dynamics_mode = 'pbd'
+    warnings.warn(
+        'dynamics_mode either not specified or not recognized, defaulting to '
+        '"pbd".  If you wish to preserve legacy behavior used in previous '
+        'versions of Brax, set dynamics_mode="legacy_spring".'
+    )
 
   # Load the meshes.
   if resource_paths is None:
     resource_paths = ['']
-  for mesh_geom in config.mesh_geometries:
-    if mesh_geom.path:
-      # Clear the vertices and faces, if any.
-      del mesh_geom.vertices[:]
-      del mesh_geom.faces[:]
-      found = False
-      for resource_path in resource_paths:
-        path = os.path.join(resource_path, mesh_geom.path)
-        if not file.Exists(path):
-          continue
-        with file.File(path, 'rb') as f:
-          trimesh = load_mesh(f, file_type=str(mesh_geom.path))
-          for v in trimesh.vertices:
-            mesh_geom.vertices.add(x=v[0], y=v[1], z=v[2])
-          mesh_geom.faces.extend(trimesh.faces.flatten())
-          for v in trimesh.vertex_normals:
-            mesh_geom.vertex_normals.add(x=v[0], y=v[1], z=v[2])
-          for v in trimesh.face_normals:
-            mesh_geom.face_normals.add(x=v[0], y=v[1], z=v[2])
-        found = True
-        break
-      assert found, f'{mesh_geom.path} is missing.'
-      mesh_geom.ClearField('path')  # Clear the path.
+  for i in range(len(config.mesh_geometries)):
+    path = config.mesh_geometries[i].path
+    if not path:
+      continue
+    mesh_geom = mesh.load(config.mesh_geometries[i].name, path, resource_paths)
+    config.mesh_geometries[i].CopyFrom(mesh_geom)
 
   # TODO: more config validation
 
