@@ -25,12 +25,23 @@ from brax.physics import bodies
 from brax.physics import colliders
 from brax.physics import config_pb2
 from brax.physics import forces
+from brax.physics import geometry
 from brax.physics import integrators
 from brax.physics import joints
 from brax.physics import spring_joints
 from brax.physics.base import Info, P, Q, QP, validate_config, vec_to_arr
 
 
+def _get_contact_info(
+    contact: Sequence[geometry.Contact]
+) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
+  if not contact:
+    return jp.array([]), jp.array([]), jp.array([])
+  contact = jp.tree_map(jp.concatenate, contact)
+  contact = jp.tree_map(lambda *x: jp.concatenate(x), *contact)
+  return contact.pos, contact.normal, contact.penetration
+
+
 @pytree.register
 class System:
   """A brax system."""
@@ -41,40 +52,53 @@ class System:
   def __init__(self,
                config: config_pb2.Config,
                resource_paths: Optional[Sequence[str]] = None):
-    self.config = validate_config(config, resource_paths=resource_paths)
-    self.num_bodies = len(config.bodies)
-    self.body = bodies.Body(config)
-    self.colliders = colliders.get(self.config, self.body)
-    self.num_joints = len(config.joints)
-    self.joints = joints.get(self.config, self.body) + spring_joints.get(
-        self.config, self.body)
+    config = validate_config(config, resource_paths=resource_paths)
+    self.config = config
     self.num_actuators = len(config.actuators)
     self.num_joint_dof = sum(len(j.angle_limit) for j in config.joints)
-    self.actuators = actuators.get(self.config, self.joints)
-    self.forces = forces.get(self.config, self.body)
+    self.num_bodies = len(config.bodies)
+    self.body = bodies.Body(config)
+    self.colliders = colliders.get(config, self.body)
+    self.num_joints = len(config.joints)
+    self.joints = joints.get(config, self.body) + spring_joints.get(
+        config, self.body)
+    self.actuators = actuators.get(config, self.joints)
+    self.forces = forces.get(config, self.body)
     self.num_forces_dof = sum(f.act_index.shape[-1] for f in self.forces)
-    self.integrator = integrators.Euler(self.config)
+    self.integrator = integrators.Euler(config)
+    zero_p = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
+    info = self.info(self.default_qp())
+    self.zero_info = Info(
+        contact=zero_p,
+        joint=zero_p,
+        actuator=zero_p,
+        contact_pos=jp.zeros_like(info.contact_pos),
+        contact_normal=jp.zeros_like(info.contact_normal),
+        contact_penetration=-jp.ones_like(info.contact_penetration))
 
   def default_angle(self, default_index: int = 0) -> jp.ndarray:
     """Returns the default joint angles for the system."""
     if not self.config.joints:
       return jp.array([])
 
-    dofs = {j.name: len(j.angle_limit) for j in self.config.joints}
+    dofs = {}
+    for j in self.config.joints:
+      dofs[j.name] = sum([l.min != 0 or l.max != 0 for l in j.angle_limit])
     angles = {}
 
     # check overrides in config defaults
     if default_index < len(self.config.defaults):
       defaults = self.config.defaults[default_index]
       for ja in defaults.angles:
-        arr = vec_to_arr(ja.angle)[:dofs[ja.name]] * jp.pi / 180
-        angles[ja.name] = arr
+        angles[ja.name] = vec_to_arr(ja.angle)[:dofs[ja.name]] * jp.pi / 180
 
     # set remaining joint angles set from angle limits, and add jitter
     for joint in self.config.joints:
       if joint.name not in angles:
-        angle = [(l.min + l.max) * jp.pi / 360 for l in joint.angle_limit]
-        angles[joint.name] = jp.array(angle)
+        dof = dofs[joint.name]
+        angles[joint.name] = jp.array([
+            (l.min + l.max) * jp.pi / 360 for l in joint.angle_limit
+        ][:dof])
 
     return jp.concatenate([angles[j.name] for j in self.config.joints])
 
@@ -106,7 +130,7 @@ class System:
     joint_idxs = []
     for j in self.config.joints:
       beg = joint_idxs[-1][1][1] if joint_idxs else 0
-      dof = len(j.angle_limit)
+      dof = sum([l.min != 0 or l.max != 0 for l in j.angle_limit])
       joint_idxs.append((j, (beg, beg + dof)))
     lineage = {j.child: j.parent for j in self.config.joints}
     depth = {}
@@ -222,6 +246,7 @@ class System:
 
   def _pbd_step(self, qp: QP, act: jp.ndarray) -> Tuple[QP, Info]:
     """Position based dynamics stepping scheme."""
+
     # Just like XPBD except performs two physics substeps per collision update.
 
     def substep(carry, _):
@@ -266,7 +291,7 @@ class System:
 
       collide_data = [c.position_apply(qp, qprev) for c in self.colliders]
       dq_c = sum([c[0] for c in collide_data], zero_q)
-      dlambda = sum([c[1] for c in collide_data])
+      dlambda = [c[1] for c in collide_data]
       contact = [c[2] for c in collide_data]
       qp = self.integrator.update(qp, pos_q=dq_c)
 
@@ -275,21 +300,20 @@ class System:
       qp = self.integrator.velocity_projection(qp, qprev)
       # apply collision velocity updates
       dp_c = sum([
-          c.velocity_apply(qp, dlambda, qp_right_before, contact[i])
+          c.velocity_apply(qp, dlambda[i], qp_right_before, contact[i])
           for i, c in enumerate(self.colliders)
       ], zero)
       qp = self.integrator.update(qp, vel_p=dp_c)
 
-      info = Info(info.contact + dp_c, info.joint, info.actuator + dp_a)
+      info = Info(info.contact + dp_c, info.joint, info.actuator + dp_a,
+                  *_get_contact_info(contact))
       return (qp, info), ()
 
     # update collider statistics for culling
     for c in self.colliders:
       c.cull.update(qp)
 
-    zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
-    info = Info(contact=zero, joint=zero, actuator=zero)
-
+    info = self.zero_info
     (qp, info), _ = jp.scan(substep, (qp, info), (), self.config.substeps // 2)
     return qp, info
 
@@ -299,13 +323,18 @@ class System:
     zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
 
     # TODO: sort out a better way to get first-step collider data
-    dq_c = sum([c.apply(qp) for c in self.colliders], zero)
+    for c in self.colliders:
+      c.cull.update(qp)
+    collider_data = [c.apply(qp) for c in self.colliders]
+    contact = [c[1] for c in collider_data]
+    dq_c = sum([c[0] for c in collider_data], zero)
     dq_j = sum([j.apply(qp) for j in self.joints], zero_q)
-    info = Info(dq_c, dq_j, zero)
+    info = Info(dq_c, dq_j, zero, *_get_contact_info(contact))
     return info
 
   def _spring_step(self, qp: QP, act: jp.ndarray) -> Tuple[QP, Info]:
     """Spring-based dynamics stepping scheme."""
+
     # Resolves actuator forces, joints, and forces at acceleration level, and
     # resolves collisions at velocity level with baumgarte stabilization.
 
@@ -323,19 +352,20 @@ class System:
       qp = self.integrator.update(qp, acc_p=dp_j + dp_a + dp_f)
 
       # apply velocity level updates for collisions
-      dp_c = sum([c.apply(qp) for c in self.colliders], zero)
+      collider_data = [c.apply(qp) for c in self.colliders]
+      dp_c = sum([c[0] for c in collider_data], zero)
+      contact = [c[1] for c in collider_data]
       qp = self.integrator.update(qp, vel_p=dp_c)
 
-      info = Info(info.contact + dp_c, info.joint + dp_j, info.actuator + dp_a)
+      info = Info(info.contact + dp_c, info.joint + dp_j, info.actuator + dp_a,
+                  *_get_contact_info(contact))
       return (qp, info), ()
 
     # update collider statistics for culling
     for c in self.colliders:
       c.cull.update(qp)
 
-    zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
-    info = Info(contact=zero, joint=zero, actuator=zero)
-
+    info = self.zero_info
     (qp, info), _ = jp.scan(substep, (qp, info), (), self.config.substeps)
     return qp, info
 
@@ -343,7 +373,11 @@ class System:
     """Return info about a system state."""
     zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
 
-    dp_c = sum([c.apply(qp) for c in self.colliders], zero)
+    for c in self.colliders:
+      c.cull.update(qp)
+    collider_data = [c.apply(qp) for c in self.colliders]
+    contact = [c[1] for c in collider_data]
+    dp_c = sum([c[0] for c in collider_data], zero)
     dp_j = sum([j.apply(qp) for j in self.joints], zero)
-    info = Info(dp_c, dp_j, zero)
+    info = Info(dp_c, dp_j, zero, *_get_contact_info(contact))
     return info
